from mindsdb.integrations.libs.const import HANDLER_TYPE

from .__about__ import __version__ as version, __description__ as description

try:
    from .anomaly_detection_handler import AnomalyDetectionHandler as Handler

    import_error = None
except Exception as e:
    Handler = None
    import_error = e

title = "Anomaly Detection"
name = "anomaly_detection"
type = HANDLER_TYPE.ML
icon_path = "icon.svg"
permanent = False

__all__ = ["Handler", "version", "name", "type", "title", "description", "import_error"]



from collections import OrderedDict

from mindsdb.integrations.libs.const import HANDLER_CONNECTION_ARG_TYPE as ARG_TYPE


connection_args = OrderedDict(
    token={
        "type": ARG_TYPE.STR,
        "description": "Box Developer Token",
        "required": True,
        "label": "Box Developer Token",
    },
)

connection_args_example = OrderedDict(
    token="",
)



from mindsdb.integrations.handlers.zipcodebase_handler.zipcodebase_tables import (
    ZipCodeBaseCodeLocationTable,
    ZipCodeBaseCodeInRadiusTable,
    ZipCodeBaseCodeByCityTable,
    ZipCodeBaseCodeByStateTable,
    ZipCodeBaseStatesByCountryTable
)
from mindsdb.integrations.handlers.zipcodebase_handler.zipcodebase import ZipCodeBaseClient
from mindsdb.integrations.libs.api_handler import APIHandler
from mindsdb.integrations.libs.response import (
    HandlerStatusResponse as StatusResponse,
)

from mindsdb.utilities import log
from mindsdb_sql_parser import parse_sql


logger = log.getLogger(__name__)


class ZipCodeBaseHandler(APIHandler):
    """The ZipCodeBase handler implementation"""

    def __init__(self, name: str, **kwargs):
        """Initialize the ZipCodeBase handler.
        Parameters
        ----------
        name : str
            name of a handler instance
        """
        super().__init__(name)

        connection_data = kwargs.get("connection_data", {})
        self.connection_data = connection_data
        self.kwargs = kwargs
        self.client = ZipCodeBaseClient(self.connection_data["api_key"])
        self.is_connected = False

        code_to_location_data = ZipCodeBaseCodeLocationTable(self)
        self._register_table("code_to_location", code_to_location_data)

        codes_within_radius_data = ZipCodeBaseCodeInRadiusTable(self)
        self._register_table("codes_within_radius", codes_within_radius_data)

        codes_by_city_data = ZipCodeBaseCodeByCityTable(self)
        self._register_table("codes_by_city", codes_by_city_data)

        codes_by_state_data = ZipCodeBaseCodeByStateTable(self)
        self._register_table("codes_by_state", codes_by_state_data)

        states_by_country_data = ZipCodeBaseStatesByCountryTable(self)
        self._register_table("states_by_country", states_by_country_data)

    def connect(self) -> StatusResponse:
        """Set up the connection required by the handler.
        Returns
        -------
        StatusResponse
            connection object
        """
        resp = StatusResponse(False)
        status = self.client.remaining_requests()
        if status["code"] != 200:
            resp.success = False
            resp.error_message = status["error"]
            return resp
        self.is_connected = True
        return resp

    def check_connection(self) -> StatusResponse:
        """Check connection to the handler.
        Returns
        -------
        StatusResponse
            Status confirmation
        """
        response = StatusResponse(False)

        try:
            status = self.client.remaining_requests()
            if status["code"] == 200:
                logger.info("Authentication successful")
                response.success = True
            else:
                response.success = False
                logger.info("Error connecting to ZipCodeBase. " + status["error"])
                response.error_message = status["error"]
        except Exception as e:
            logger.error(f"Error connecting to ZipCodeBase: {e}!")
            response.error_message = e

        self.is_connected = response.success
        return response

    def native_query(self, query: str) -> StatusResponse:
        """Receive and process a raw query.
        Parameters
        ----------
        query : str
            query in a native format
        Returns
        -------
        StatusResponse
            Request status
        """
        ast = parse_sql(query)
        return self.query(ast)



import unittest
from mindsdb.integrations.handlers.firebird_handler.firebird_handler import FirebirdHandler
from mindsdb.api.executor.data_types.response_type import RESPONSE_TYPE


class FirebirdHandlerTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.kwargs = {
            "connection_data": {
                "host": "localhost",
                "database": r"C:\Users\minura\Documents\mindsdb\test.fdb",
                "user": "sysdba",
                "password": "password"
            }
        }
        cls.handler = FirebirdHandler('test_firebird_handler', cls.kwargs)

    def test_0_check_connection(self):
        assert self.handler.check_connection()

    def test_1_native_query_select(self):
        query = "SELECT * FROM test_tbl"
        result = self.handler.native_query(query)
        assert result.type is RESPONSE_TYPE.TABLE

    def test_2_get_tables(self):
        tables = self.handler.get_tables()
        assert tables.type is not RESPONSE_TYPE.ERROR

    def test_4_get_columns(self):
        columns = self.handler.get_columns('test_tbl')
        assert columns.type is not RESPONSE_TYPE.ERROR


if __name__ == '__main__':
    unittest.main()



from mindsdb.integrations.libs.const import HANDLER_TYPE
from .__about__ import __version__ as version, __description__ as description
from mindsdb.utilities import log


logger = log.getLogger(__name__)

try:
    from .google_gemini_handler import GoogleGeminiHandler as Handler

    import_error = None
except Exception as e:
    Handler = None
    import_error = e

title = "Google Gemini"
name = "google_gemini"
type = HANDLER_TYPE.ML
icon_path = "icon.svg"
permanent = False
__all__ = [
    "Handler",
    "version",
    "name",
    "type",
    "title",
    "description",
    "import_error",
    "icon_path",
]




from typing import Union

from mindsdb_sql_parser.ast import Identifier, Select, BinaryOperation, Constant, OrderBy

from mindsdb.interfaces.storage import db
from .types import ChatBotMessage


class BaseMemory:
    '''
    base class to work with chatbot memory
    '''
    MAX_DEPTH = 100

    def __init__(self, chat_task, chat_params):
        # in memory yet
        self._modes = {}
        self._hide_history_before = {}
        self._cache = {}
        self.chat_params = chat_params
        self.chat_task = chat_task

    def get_chat(self, chat_id, table_name=None):
        return ChatMemory(self, chat_id, table_name=table_name)

    def hide_history(self, chat_id, left_count, table_name=None):
        '''
        set date to start hiding messages
        '''
        history = self.get_chat_history(chat_id, table_name=table_name)
        if left_count > len(history) - 1:
            left_count = len(history) - 1
        sent_at = history[-left_count].sent_at

        self._hide_history_before[chat_id] = sent_at

    def _apply_hiding(self, chat_id, history):
        '''
        hide messages from history
        '''
        before = self._hide_history_before.get(chat_id)

        if before is None:
            return history

        return [
            msg
            for msg in history
            if msg.sent_at >= before
        ]

    def get_mode(self, chat_id):
        return self._modes.get(chat_id)

    def set_mode(self, chat_id, mode):
        self._modes[chat_id] = mode

    def add_to_history(self, chat_id, chat_message, table_name=None):

        # If the chat_id is a tuple, convert it to a string when storing the message in the database.
        self._add_to_history(
            chat_id,
            chat_message,
            table_name=table_name
        )
        if chat_id in self._cache:
            del self._cache[chat_id]

    def get_chat_history(self, chat_id, table_name=None, cached=True):
        key = (chat_id, table_name) if table_name else chat_id
        if cached and key in self._cache:
            history = self._cache[key]

        else:
            history = self._get_chat_history(
                chat_id,
                table_name
            )
            self._cache[key] = history

        history = self._apply_hiding(chat_id, history)
        return history

    def _add_to_history(self, chat_id, chat_message, table_name=None):
        raise NotImplementedError

    def _get_chat_history(self, chat_id, table_name=None):
        raise NotImplementedError


class HandlerMemory(BaseMemory):
    '''
    Uses handler's database to store and retrieve messages
    '''

    def _add_to_history(self, chat_id, chat_message, table_name=None):
        # do nothing. sent message will be stored by handler db
        pass

    def _get_chat_history(self, chat_id, table_name):
        t_params = next(
            chat_params['chat_table'] for chat_params in self.chat_params if chat_params['chat_table']['name'] == table_name
        )

        text_col = t_params['text_col']
        username_col = t_params['username_col']
        time_col = t_params['time_col']
        chat_id_cols = t_params['chat_id_col'] if isinstance(t_params['chat_id_col'], list) else [t_params['chat_id_col']]

        chat_id = chat_id if isinstance(chat_id, tuple) else (chat_id,)
        # Add a WHERE clause for each chat_id column.
        where_conditions = [
            BinaryOperation(
                op='=',
                args=[
                    Identifier(chat_id_col),
                    Constant(chat_id[idx])
                ]
            ) for idx, chat_id_col in enumerate(chat_id_cols)
        ]
        # Add a WHERE clause to ignore holding messages from the bot.
        from .chatbot_task import HOLDING_MESSAGE

        where_conditions.append(
            BinaryOperation(
                op='!=',
                args=[
                    Identifier(text_col),
                    Constant(HOLDING_MESSAGE)
                ]
            )
        )

        # Convert the WHERE conditions to a BinaryOperation object.
        where_conditions_binary_operation = None
        for condition in where_conditions:
            if where_conditions_binary_operation is None:
                where_conditions_binary_operation = condition
            else:
                where_conditions_binary_operation = BinaryOperation('and', args=[where_conditions_binary_operation, condition])

        ast_query = Select(
            targets=[Identifier(text_col),
                     Identifier(username_col),
                     Identifier(time_col)],
            from_table=Identifier(t_params['name']),
            where=where_conditions_binary_operation,
            order_by=[OrderBy(Identifier(time_col))],
            limit=Constant(self.MAX_DEPTH),
        )

        resp = self.chat_task.chat_handler.query(ast_query)
        if resp.data_frame is None:
            return

        df = resp.data_frame

        # get last messages
        df = df.iloc[-self.MAX_DEPTH:]

        result = []
        for _, rec in df.iterrows():
            chatbot_message = ChatBotMessage(
                ChatBotMessage.Type.DIRECT,
                rec[text_col],
                user=rec[username_col],
                sent_at=rec[time_col]
            )
            result.append(chatbot_message)

        return result


class DBMemory(BaseMemory):
    '''
    uses mindsdb database to store messages
    '''

    def _generate_chat_id_for_db(self, chat_id: Union[str, tuple], table_name: str = None) -> str:
        """
        Generate an ID for the chat to store in the database.
        The ID is a string that includes the components of the chat ID and the table name (if provided) separated by underscores.

        Args:
            chat_id (str | tuple): The ID of the chat.
            table_name (str): The name of the table the chat belongs to.
        """
        if isinstance(chat_id, tuple):
            char_id_str = "_".join(str(val) for val in chat_id)
        else:
            char_id_str = str(chat_id)

        if table_name:
            chat_id_str = f"{table_name}_{char_id_str}"

        return chat_id_str

    def _add_to_history(self, chat_id, message, table_name=None):
        chat_bot_id = self.chat_task.bot_id
        destination = self._generate_chat_id_for_db(chat_id, table_name)

        message = db.ChatBotsHistory(
            chat_bot_id=chat_bot_id,
            type=message.type.name,
            text=message.text,
            user=message.user,
            destination=destination,
        )
        db.session.add(message)
        db.session.commit()

    def _get_chat_history(self, chat_id, table_name=None):
        chat_bot_id = self.chat_task.bot_id
        destination = self._generate_chat_id_for_db(chat_id, table_name)

        query = db.ChatBotsHistory.query\
            .filter(
                db.ChatBotsHistory.chat_bot_id == chat_bot_id,
                db.ChatBotsHistory.destination == destination
            )\
            .order_by(db.ChatBotsHistory.sent_at.desc())\
            .limit(self.MAX_DEPTH)

        result = [
            ChatBotMessage(
                rec.type,
                rec.text,
                rec.user,
                sent_at=rec.sent_at,
            )
            for rec in query
        ]
        result.reverse()
        return result


class ChatMemory:
    '''
    interface to work with individual chat
    '''
    def __init__(self, memory, chat_id, table_name=None):
        self.memory = memory
        self.chat_id = chat_id
        self.table_name = table_name

        self.cached = False

    def get_history(self, cached=True):
        result = self.memory.get_chat_history(self.chat_id, self.table_name, cached=cached and self.cached)

        self.cached = True
        return result

    def add_to_history(self, message):
        self.memory.add_to_history(self.chat_id, message, table_name=self.table_name)

    def get_mode(self):
        return self.memory.get_mode(self.chat_id)

    def set_mode(self, mode):
        self.memory.set_mode(self.chat_id, mode)

    def hide_history(self, left_count):
        '''
        set date to start hiding messages
        '''
        self.memory.hide_history(self.chat_id, left_count, table_name=self.table_name)



import json

import requests
from typing import Dict, Optional

import pandas as pd

from mindsdb.integrations.libs.base import BaseMLEngine


class RayServeException(Exception):
    pass


class RayServeHandler(BaseMLEngine):
    """
    The Ray Serve integration engine needs to have a working connection to Ray Serve. For this:
        - A Ray Serve server should be running

    Example:

    """  # noqa
    name = 'ray_serve'

    @staticmethod
    def create_validation(target, args=None, **kwargs):
        if not args.get('using'):
            raise Exception("Error: This engine requires some parameters via the 'using' clause. Please refer to the documentation of the Ray Serve handler and try again.")  # noqa
        if not args['using'].get('train_url'):
            raise Exception("Error: Please provide a URL for the training endpoint.")
        if not args['using'].get('predict_url'):
            raise Exception("Error: Please provide a URL for the prediction endpoint.")

    def create(self, target: str, df: Optional[pd.DataFrame] = None, args: Optional[Dict] = None) -> None:
        # TODO: use join_learn_process to notify users when ray has finished the training process
        args = args['using']  # ignore the rest of the problem definition
        args['target'] = target
        self.model_storage.json_set('args', args)
        try:
            resp = requests.post(args['train_url'],
                                 json={'df': df.to_json(orient='records'), 'target': target},
                                 headers={'content-type': 'application/json; format=pandas-records'})
        except requests.exceptions.InvalidSchema:
            raise Exception("Error: The URL provided for the training endpoint is invalid.")

        error = None
        try:
            resp = resp.json()
        except json.JSONDecodeError:
            error = resp.text
        else:
            if resp.get('status') != 'ok':
                error = resp['status']

        if error:
            raise RayServeException(f"Error: {error}")

    def predict(self, df, args=None):
        args = {**(self.model_storage.json_get('args')), **args}  # merge incoming args
        pred_args = args.get('predict_params', {})
        args = {**args, **pred_args}  # merge pred_args
        resp = requests.post(args['predict_url'],
                             json={'df': df.to_json(orient='records'), 'pred_args': pred_args},
                             headers={'content-type': 'application/json; format=pandas-records'})

        try:
            response = resp.json()
        except json.JSONDecodeError:
            error = resp.text
        else:
            if 'prediction' in response:
                target = args['target']
                if target != 'prediction':
                    # rename prediction to target
                    response[target] = response.pop('prediction')
                return pd.DataFrame(response)
            else:
                # something wrong
                error = response

        raise RayServeException(f"Error: {error}")

    def describe(self, key: Optional[str] = None) -> pd.DataFrame:
        args = self.model_storage.json_get('args')
        description = {
            'TRAIN_URL': [args['train_url']],
            'PREDICT_URL': [args['predict_url']],
            'TARGET': [args['target']],
        }
        return pd.DataFrame.from_dict(description)



import stripe
from mindsdb.integrations.handlers.stripe_handler.stripe_tables import CustomersTable, ProductsTable, PaymentIntentsTable, RefundsTable, PayoutsTable
from mindsdb.integrations.libs.api_handler import APIHandler
from mindsdb.integrations.libs.response import (
    HandlerStatusResponse as StatusResponse,
)

from mindsdb.utilities import log
from mindsdb_sql_parser import parse_sql

logger = log.getLogger(__name__)

class StripeHandler(APIHandler):
    """
    The Stripe handler implementation.
    """

    name = 'stripe'

    def __init__(self, name: str, **kwargs):
        """
        Initialize the handler.
        Args:
            name (str): name of particular handler instance
            **kwargs: arbitrary keyword arguments.
        """
        super().__init__(name)

        connection_data = kwargs.get("connection_data", {})
        self.connection_data = connection_data
        self.kwargs = kwargs

        self.connection = None
        self.is_connected = False

        customers_data = CustomersTable(self)
        self._register_table("customers", customers_data)

        products_data = ProductsTable(self)
        self._register_table("products", products_data)

        payment_intents_data = PaymentIntentsTable(self)
        self._register_table("payment_intents", payment_intents_data)
        
        payouts_data = PayoutsTable(self)
        self._register_table("payouts", payouts_data)

        refunds_data = RefundsTable(self)
        self._register_table("refunds", refunds_data)

    def connect(self):
        """
        Set up the connection required by the handler.
        Returns
        -------
        StatusResponse
            connection object
        """
        if self.is_connected is True:
            return self.connection

        stripe.api_key = self.connection_data['api_key']

        self.connection = stripe
        self.is_connected = True

        return self.connection

    def check_connection(self) -> StatusResponse:
        """
        Check connection to the handler.
        Returns:
            HandlerStatusResponse
        """

        response = StatusResponse(False)

        try:
            stripe = self.connect()
            stripe.Account.retrieve()
            response.success = True
        except Exception as e:
            logger.error(f'Error connecting to Stripe!')
            response.error_message = str(e)

        self.is_connected = response.success

        return response

    def native_query(self, query: str) -> StatusResponse:
        """Receive and process a raw query.
        Parameters
        ----------
        query : str
            query in a native format
        Returns
        -------
        StatusResponse
            Request status
        """
        ast = parse_sql(query)
        return self.query(ast)



from mindsdb.api.mongo.classes import Responder
from mindsdb.utilities.context import context as ctx


class Responce(Responder):
    def when(self, query):
        return 'company_id' in query

    def result(self, query, request_env, mindsdb_env, session):
        company_id = query.get('company_id')
        user_class = query.get('user_class', 0)
        email_confirmed = query.get('email_confirmed', 1)
        need_response = query.get('need_response', False)

        ctx.company_id = company_id
        ctx.user_class = user_class
        ctx.email_confirmed = email_confirmed

        if need_response:
            return {'ok': 1}

        return None


responder = Responce()
